package bin;

import java.io.File;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;

import shared.Parse;
import shared.Parser;
import shared.Timer;
import shared.Tools;
import tax.TaxTree;

public class CladeIndex {
	
	public static void main(String[] args) {
		Timer t=new Timer();

		//Create a parser object
		Parser parser=new Parser();
		
		PrintStream outstream=System.out;
		ArrayList<String> in=new ArrayList<String>();
		ArrayList<String> ref=new ArrayList<String>();
		
		//Parse each argument
		for(int i=0; i<args.length; i++){
			String arg=args[i];
			
			//Break arguments into their constituent parts, in the form of "a=b"
			String[] split=arg.split("=");
			String a=split[0].toLowerCase();
			String b=split.length>1 ? split[1] : null;
			if(b!=null && b.equalsIgnoreCase("null")){b=null;}
			
			if(a.equals("steps")){
				maxSteps=Integer.parseInt(b);
			}else if(a.equalsIgnoreCase("comparisonCutoffMult") || a.equals("ccm")){
				Comparison.setComparisonCutoffMult(Float.parseFloat(b));
			}else if(a.equals("gcdelta") || a.equals("gcdif")){
				gcDelta=Float.parseFloat(b);
			}else if(a.equals("abs") || a.equals("absdif")){
				Comparison.method=Comparison.ABS;
			}else if(a.equals("cos") || a.equals("cosine")){
				Comparison.method=Comparison.COS;
			}else if(a.equals("hel") || a.equals("hellinger")){
				Comparison.method=Comparison.HEL;
			}else if(a.equals("euc") || a.equals("euclidian")){
				Comparison.method=Comparison.EUC;
			}else if(a.equals("usetree")){
				useTree=Parse.parseBoolean(b);
			}else if(a.equals("printmetrics")){
				printMetrics=Parse.parseBoolean(b);
			}else if(a.equals("banself")){
				banSelf=Parse.parseBoolean(b);
			}else if(a.equals("includeself")){
				banSelf=!Parse.parseBoolean(b);
			}else if(a.equals("ref")){
				Tools.getFileOrFiles(b, ref, true, false, false, false);
			}else if(a.equals("ref")){
				Tools.getFileOrFiles(b, ref, true, false, false, false);
			}else if(a.equals("in")){
				Tools.getFileOrFiles(b, in, true, false, false, false);
			}else if(b==null && new File(arg).isFile()){
				in.add(arg);
			}else if(b==null && new File(arg).isDirectory()){
				Tools.getFileOrFiles(arg, in, true, false, false, false);
			}else if(a.equals("parse_flag_goes_here")){
				long fake_variable=Parse.parseKMG(b);
				//Set a variable here
			}else if(parser.parse(arg, a, b)){//Parse standard flags in the parser
				//do nothing
			}else{
				outstream.println("Unknown parameter "+args[i]);
				assert(false) : "Unknown parameter "+args[i];
			}
		}
		
		if(useTree) {BinObject.loadTree();}
		CladeLoader loader=new CladeLoader();
		ConcurrentHashMap<Integer, Clade> map=loader.load(ref, null);
		CladeIndex index=new CladeIndex(map.values());
		t.stopAndStart("Indexed "+map.size()+" clades in ");
		map=null;
		index.search(in);
		t.stop("\nSearched "+in.size()+" queries and made "+index.comparisons+" comparisons in ");
	}
	
	private void search(ArrayList<String> queries) {
		int correct=0, wrong=0, nohit=0, other=0;
		int[] levels=new int[TaxTree.numTaxLevelNames];
		
		double[][] metrics=new double[7][TaxTree.numTaxLevelNames];
		String[] mNames= {"r34", "r45", "strand", "gc", "3mer", "4mer", "5mer"};
		double[] ratio34=metrics[0];
		double[] ratio45=metrics[1];
		double[] strandDif=metrics[2];
		double[] gcDif=metrics[3];
		double[] triDif=metrics[4];
		double[] tetDif=metrics[5];
		double[] pentDif=metrics[6];
		
		for(String fname : queries) {
			Comparison c=search(fname);
			System.err.println();
			if(c==null || c.ref==null) {nohit++;}
			else if(c.correct()) {correct++;}
			else if(c.incorrect()) {wrong++;}
			else {other++;}
			int level=levels.length-1;
			
			if(useTree) {
				level=c.correctLevel();
				levels[level]++;
			}
			levels[level]++;
			
			if(c!=null && c.ref!=null) {
				ratio34[level]+=c.k3dif/Math.max(c.k4dif, 0.000001f);
				ratio45[level]+=c.k4dif/Math.max(c.k5dif, 0.000001f);
				strandDif[level]+=c.strdif;
				gcDif[level]+=c.gcdif;
				triDif[level]+=c.k3dif;
				tetDif[level]+=c.k4dif;
				pentDif[level]+=c.k5dif;
			}
		}
		System.err.println("Results:\n");
		System.err.println("Correct:  \t"+correct);
		System.err.println("Incorrect:\t"+wrong);
		System.err.println("No-Hit:   \t"+nohit);
		System.err.println("Other:    \t"+other);
		
		if(useTree) {
			System.err.println("\nBest Hit Accuracy:");
			long score=0, sum=0;
			for(int level=0; level<levels.length; level++) {
				int x=levels[level];
				if(x>0) {
					sum+=x;
					score+=(levels.length-level-1);
					System.err.println(Tools.padRight(TaxTree.levelToString(level)+":", 9)+"\t"+x);
				}
			}
			System.err.println("\n"+Tools.padRight("Score:", 9)+String.format("\t%.2f", score/(float)sum));
		}

		if(printMetrics) {
			System.err.println("\nMetrics:");
			for(int level=0; level<levels.length; level++) {System.err.print(TaxTree.levelToStringShort(level)+"\t");}
			System.err.println();
			for(int metric=0; metric<metrics.length; metric++) {
				System.err.print(mNames[metric]);
				for(int level=0; level<levels.length; level++) {
					int count=levels[level];
//					if(count==0) {System.err.print(0);}
					System.err.print(String.format("\t%.6f", metrics[metric][level]/count));
				}
				System.err.println();
			}
		}
	}
	
	Comparison search(String fname) {
		Clade query=CladeLoader.loadOneClade(fname);
		Comparison best=findBest(query);
		System.err.println("Query:     \t"+query);
		System.err.println("Best Match:\t"+best);
		return best;
	}
	
	public CladeIndex(Collection<Clade> coll) {
		for(Clade c : coll) {add(c);}
		System.err.println("Indexed "+coll.size()+" clades.");
	}
	
	public void add(Clade c) {
		int idx=Math.round(c.gc*100);
		ArrayList<Clade> list=gcDex[idx];
		if(list==null) {gcDex[idx]=list=new ArrayList<Clade>(16);}
		list.add(c);
	}
	
	private Comparison findBest(Clade c) {
//		System.err.println("Searching for best match for "+c);
		int center=Math.round(c.gc*100);
		Comparison temp=new Comparison();
		Comparison best=findBest(c, gcDex[center], null, temp);
		for(int i=1; i<=maxSteps; i++) {
			int low=center-i, high=center+i;
			if(low>=0) {best=findBest(c, gcDex[low], best, temp);}
			if(high<gcDex.length) {best=findBest(c, gcDex[high], best, temp);}
		}
		return best;
	}
	
	public Comparison findBest(Clade a, ArrayList<Clade> list, Comparison best, Comparison temp) {
		if(list==null || list.isEmpty()) {return best;}
//		System.err.println("\nSearching a list of size "+list.size());
		if(best==null) {best=new Comparison();}
		float maxgcdif=best.gcdif+gcDelta;
		float max5merdif=best.k5dif;
		for(Clade b : list) {
			if(b==a || (b.taxID==a.taxID && banSelf)) {continue;}
//			System.err.println("Comparing to "+b);
			temp.compare(a, b, maxgcdif, max5merdif);
			comparisons++;
//			System.err.println("Comparison: "+temp);
			int x=temp.compareTo(best);
			if(x<0) {
//				System.err.println("***New best!");
				best.setFrom(temp);
				maxgcdif=best.gcdif+gcDelta;
				max5merdif=best.k5dif;
			}
		}
		return best;
	}
	
	@SuppressWarnings("unchecked")
	final ArrayList<Clade>[] gcDex=new ArrayList[101];
	
	long comparisons=0;

	static boolean useTree=false;
	static boolean printMetrics=false;
	static boolean banSelf=false;
	static int maxSteps=5;
	static float gcDelta=0.05f;
	
}
